

import collections, asyncio, sys

import xml.etree.ElementTree as ET

from functools import partialmethod

import logging
logger = logging.getLogger(__name__)

from .ipydriver import IPyDriver

from .comms import Port_RX, Port_TX, cleanque, SendChecker, TXTimer

from .remote import RemoteConnection

from .exdriver import ExDriver


class IPyServer:

    """Once an instance of this class is created, the asyncrun method
       should be awaited which will open a port, and the INDI service
       will be available for clients to connect.

       drivers is a list of IPyDriver objects this driver handles,
       host and port are "localhost" and 7624 as default

       maxconnections is the number of simultaneous client connections
       accepted, with a default of 5. The number given should be
       between 1 and 10 inclusive.

       If, prior to asyncrun being awaited, the add_remote method is called,
       then a connection will be made to a remote INDI server and any of its
       drivers.

       The add_remote method can be called multiple times to create
       connections to different servers making a branching
       tree of servers and drivers.

       The add_exdriver method can be called to run an executable driver and
       this server will communicate to it by stdin, stdout and stderr, therefore
       ipyserver can act as a general INDI server for third party drivers as
       well as ipydriver instances.
       """


    def __init__(self, drivers, *, host="localhost", port=7624, maxconnections=5):


        if isinstance(drivers, IPyDriver):
            # Its a single IPyDriver rather than a list
            self.drivers = [drivers]
        else:
            self.drivers = drivers

        self.host = host
        self.port = port

        # traffic is transmitted out on the serverwriterque
        self.serverwriterque = asyncio.Queue(6)
        # and read in from the serverreaderque
        self.serverreaderque = asyncio.Queue(6)

        if maxconnections<1 or maxconnections>10:
            raise ValueError("maxconnections should be a number between 1 and 10 inclusive.")
        self.maxconnections = maxconnections

        # this is a dictionary of device name to device
        self.devices = {}

        # self.remotes is a list of RemoteConnection objects running connections to remote servers
        # this list is populated by calling self.add_remote(host, port, debug_enable)
        self.remotes = []

        # self.exdrivers is a list of ExDriver objects running external drivers
        # this list is populated by calling self.add_exdriver(program, *args, debug_enable=False)
        self.exdrivers = []

        for driver in self.drivers:
            if not isinstance(driver, IPyDriver):
                raise TypeError("The drivers set in IPyServer must all be IPyDrivers")
            if not driver.comms is None:
                 raise RuntimeError("A driver communications method has already been set, there can only be one")
            devicesindriver = driver.devices.copy()
            for devicename in devicesindriver:
                if devicename in self.devices:
                    # duplicate devicename
                    raise ValueError(f"Device name {devicename} is duplicated in the attached drivers.")
            self.devices.update(devicesindriver)

        self.connectionpool = []
        for clientconnection in range(0, maxconnections):
            self.connectionpool.append(_ClientConnection(self.devices, self.exdrivers, self.remotes, self.serverreaderque))

        # This alldrivers list will have exdrivers added to it, so the list
        # here is initially a copy of self.drivers
        self.alldrivers = self.drivers.copy()

        for driver in self.drivers:
            # an instance of _DriverComms is created for each driver
            # each _DriverComms object has lists of drivers and remotes
            # these will be used to send snooping traffic

            driver.comms = _DriverComms(driver,
                                        self.serverwriterque,
                                        self.connectionpool,
                                        self.alldrivers,
                                        self.remotes)
        # shutdown routine sets this to True to stop coroutines
        self._stop = False
        self.server = None

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the server"
        return self._stop

    def shutdown(self):
        "Shuts down the server, sets the flag self._stop to True"
        self._stop = True
        for driver in self.drivers:
            driver.shutdown()
        for remcon in self.remotes:
            remcon.shutdown()
        for exd in self.exdrivers:
            exd.shutdown()
        for clientconnection in self.connectionpool:
            clientconnection.shutdown()
        if not self.server is None:
            self.server.close()


    def add_remote(self, host, port, blob_enable="Never", debug_enable=False):
        """Adds a connection to a remote server.
           blob_enable can be Never, Also or Only.
           If Never BLOBs will not be sent from the remote server to this one.
           If Also BLOBs and other vectors can all be sent.
           If Only, then only BLOB traffic will be sent.

           If debug_enable is True, then DEBUG level logging will record xml
           traffic, if False, the xml traffic will not be logged. This can be
           used to prevent multiple such connections all logging xml traffic together."""


        snoopall = False           # gets set to True if it is snooping everything
        snoopdevices = set()       # gets set to a set of device names
        snoopvectors = set()       # gets set to a set of (devicename,vectorname) tuples

        remcon = RemoteConnection(indihost=host, indiport=port,
                                  alldrivers = self.alldrivers,
                                  remotes = self.remotes,
                                  serverwriterque = self.serverwriterque,
                                  connectionpool = self.connectionpool,
                                  blob_enable = blob_enable,
                                  snoopall = snoopall,
                                  snoopdevices = snoopdevices,
                                  snoopvectors = snoopvectors )

        if debug_enable:
            remcon.debug_verbosity(2) # turn on xml logs
        else:
            remcon.debug_verbosity(0) # turn off xml logs
        # turn off timers, these are more appropriate to a client
        remcon.set_vector_timeouts(timeout_enable=False)
        # store this object
        self.remotes.append(remcon)


    def add_exdriver(self, program, *args, debug_enable=False):
        """Adds an executable driver program, runs it and communicates to it via stdin, stdout
           and stderr. Then serves the driver, and any others added, by the listening port.
           args is used for the program arguments if any.
           Any program output on stderr will be logged at level ERROR.

           If debug_enable is True, then DEBUG level logging will record xml
           traffic, if False, the xml traffic will not be logged. This can be
           used to prevent multiple drivers all logging xml traffic together."""
        exd = ExDriver(program, *args, debug_enable=debug_enable)
        # add this exdriver to alldrivers
        self.alldrivers.append(exd)
        # Create a DriverComms object
        exd.comms = _DriverComms(exd,
                                 self.serverwriterque,
                                 self.connectionpool,
                                 self.alldrivers,
                                 self.remotes)
        # store this object
        self.exdrivers.append(exd)



    async def _runserver(self):
        "Runs the server on the given host and port"
        logger.info(f"{self.__class__.__name__} listening on {self.host} : {self.port}")
        self.server = await asyncio.start_server(self.handle_data, self.host, self.port)
        async with self.server:
            await self.server.serve_forever()


    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"

        for clientconnection in self.connectionpool:
            if not clientconnection.connected:
                # this clientconnection is available
                await clientconnection.handle_data(reader, writer)
                break
        else:
            # no clientconnection is available
            writer.close()
            await writer.wait_closed()


    async def asyncrun(self):
        """Runs the server together with its drivers and any remote connections."""
        driverruns = [ driver.asyncrun() for driver in self.drivers ]
        remoteruns = [ remoteconnection.asyncrun() for remoteconnection in self.remotes ]
        externalruns = [ exd.asyncrun() for exd in self.exdrivers ]
        await asyncio.gather(*driverruns,
                             *remoteruns,
                             *externalruns,
                             self._runserver(),
                             self._copyfromserver(),
                             self._sendtoclient()
                             )


    async def _copyfromserver(self):
        """Gets data from serverreaderque.
           For every driver, copy data, if applicable, to driver.readerque
           And for every remote connection if applicable, to its send method"""
        while not self._stop:
            await asyncio.sleep(0)
            xmldata = await self.serverreaderque.get()
            devicename = xmldata.get("device")
            propertyname = xmldata.get("name")

            remconfound = False
            exdriverfound = False

            # check for a getProperties
            if xmldata.tag == "getProperties":
                # if getproperties is targetted at a known device, send it to that device
                if devicename:
                    if devicename in self.devices:
                        # this getProperties request is meant for an attached device
                        await self.devices[devicename].driver.readerque.put(xmldata)
                        # no need to transmit this anywhere else, continue the while loop
                        self.serverreaderque.task_done()
                        continue
                    for remcon in self.remotes:
                        if devicename in remcon:
                            # this getProperties request is meant for a remote connection
                            remcon.send(xmldata)
                            remconfound = True
                            break
                    if not remconfound:
                        for exd in self.exdrivers:
                            if devicename in exd:
                                # this getProperties request is meant for an external driver
                                await exd.readerque.put(xmldata)
                                exdriverfound = True
                                break

            if remconfound:
                # no need to transmit this anywhere else, continue the while loop
                self.serverreaderque.task_done()
                continue

            if exdriverfound:
                # no need to transmit this anywhere else, continue the while loop
                self.serverreaderque.task_done()
                continue


            # transmit xmldata out to remote connections
            if xmldata.tag != "enableBLOB":
                # enableBLOB instructions are not forwarded to remcon's
                for remcon in self.remotes:
                    if not remcon.connected:
                        continue
                    if devicename and (devicename in remcon):
                        # this devicename has been found on this remote,
                        # so it must be a 'new' intended for this connection and
                        # it is not snoopable, since it is data to a device, not from it.
                        remcon.send(xmldata)
                        remconfound = True
                        break
                    elif xmldata.tag == "getProperties":
                        # either no devicename, or an unknown device
                        # if it were a known devicename the previous block would have handled it.
                        # so send it on all connections
                        remcon.send(xmldata)
                    elif not xmldata.tag.startswith("new"):
                        # either devicename is unknown, or this data is to/from another driver.
                        # So check if this remcon is snooping on this device/vector
                        # only forward def's and set's, not 'new' vectors which
                        # do not come from a device, but only from a client to the target device.
                        if remcon.clientdata["snoopall"]:
                            remcon.send(xmldata)
                        elif devicename and (devicename in remcon.clientdata["snoopdevices"]):
                            remcon.send(xmldata)
                        elif devicename and propertyname and ((devicename, propertyname) in remcon.clientdata["snoopvectors"]):
                            remcon.send(xmldata)

            if remconfound:
                # no need to transmit this anywhere else, continue the while loop
                self.serverreaderque.task_done()
                continue

            # transmit xmldata out to exdrivers
            if xmldata.tag != "enableBLOB":
                # enableBLOB instructions are not forwarded to external drivers
                for driver in self.exdrivers:
                    if devicename and (devicename in driver):
                        # data is intended for this driver
                        # it is not snoopable, since it is data to a device, not from it.
                        await driver.readerque.put(xmldata)
                        exdriverfound = True
                        break
                    elif xmldata.tag == "getProperties":
                        # either no devicename, or an unknown device
                        await driver.readerque.put(xmldata)
                    elif not xmldata.tag.startswith("new"):
                        # either devicename is unknown, or this data is to/from another driver.
                        # So check if this driver is snooping on this device/vector
                        # only forward def's and set's, not 'new' vectors which
                        # do not come from a device, but only from a client to the target device.
                        if driver.snoopall:
                            await driver.readerque.put(xmldata)
                        elif devicename and (devicename in driver.snoopdevices):
                            await driver.readerque.put(xmldata)
                        elif devicename and propertyname and ((devicename, propertyname) in driver.snoopvectors):
                            await driver.readerque.put(xmldata)

            if exdriverfound:
                # no need to transmit this anywhere else, continue the while loop
                self.serverreaderque.task_done()
                continue

            # transmit xmldata out to drivers
            for driver in self.drivers:
                if devicename and (devicename in driver):
                    # data is intended for this driver
                    # it is not snoopable, since it is data to a device, not from it.
                    await driver.readerque.put(xmldata)
                    break
                elif xmldata.tag == "getProperties":
                    # either no devicename, or an unknown device
                    await driver.readerque.put(xmldata)
                elif not xmldata.tag.startswith("new"):
                    # either devicename is unknown, or this data is to/from another driver.
                    # So check if this driver is snooping on this device/vector
                    # only forward def's and set's, not 'new' vectors which
                    # do not come from a device, but only from a client to the target device.
                    if driver.snoopall:
                        await driver.readerque.put(xmldata)
                    elif devicename and (devicename in driver.snoopdevices):
                        await driver.readerque.put(xmldata)
                    elif devicename and propertyname and ((devicename, propertyname) in driver.snoopvectors):
                        await driver.readerque.put(xmldata)

            self.serverreaderque.task_done()
            # now every driver/remcon which needs it has this xmldata



    async def _sendtoclient(self):
        "For every clientconnection, get txque and copy data into it from serverwriterque"
        while not self._stop:
            await asyncio.sleep(0)
            xmldata = await self.serverwriterque.get()
            for clientconnection in self.connectionpool:
                if clientconnection.connected:
                    await clientconnection.txque.put(xmldata)
            # task completed
            self.serverwriterque.task_done()



class _DriverComms:

    """An instance of this is created for each driver, which calls the __call__
       method.  Any data the driver wishes to be send will be taken
       from the drivers writerque and transmitted to the client by placing it
       into the serverwriterque"""

    def __init__(self, driver, serverwriterque, connectionpool, alldrivers, remotes):

        # This object is attached to this driver
        self.driver = driver
        self.serverwriterque = serverwriterque
        # connectionpool is a list of ClientConnection objects, which is used
        # to test if a client is connected
        self.connectionpool = connectionpool
        # self.connected is read by the driver, and in this case is always True
        # as the driver is connected to IPyServer, which handles snooping traffic,
        # even if no client is connected
        self.connected = True
        # self.alldrivers is set to a list of drivers, including exdrivers
        self.alldrivers = alldrivers
        # self.remotes is a list of connections to remote servers
        self.remotes = remotes
        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        "Sets self.stop to True and calls shutdown on tasks"
        self._stop = True


    async def __call__(self, readerque, writerque):
        """Called by the driver, should run continuously.
           reads writerque from the driver, and sends xml data to the network"""
        while not self._stop:
            await asyncio.sleep(0)
            xmldata = await writerque.get()
            # Check if other drivers/remotes wants to snoop this traffic
            devicename = xmldata.get("device")
            propertyname = xmldata.get("name")

            if xmldata.tag.startswith("new"):
                # drivers should never transmit a new
                # but just in case
                writerque.task_done()
                logger.error(f"Driver transmitted invalid tag {xmldata.tag}")
                continue

            # check for a getProperties
            if xmldata.tag == "getProperties":
                foundflag = False
                # if getproperties is targetted at a known device, send it to that device
                if devicename:
                    for driver in self.alldrivers:
                        if driver is self.driver:
                            # No need to check sending a getProperties to itself
                            continue
                        if devicename in driver:
                            # this getProperties request is meant for an attached driver/device
                            await driver.readerque.put(xmldata)
                            foundflag = True
                            break
                    if foundflag:
                        # no need to transmit this anywhere else, continue the while loop
                        writerque.task_done()
                        continue
                    for remcon in self.remotes:
                        if not remcon.connected:
                            continue
                        if devicename in remcon:
                            # this getProperties request is meant for a remote connection
                            remcon.send(xmldata)
                            foundflag = True
                            break
                    if foundflag:
                        # no need to transmit this anywhere else, continue the while loop
                        writerque.task_done()
                        continue

            # transmit xmldata out to remote connections
            for remcon in self.remotes:
                if xmldata.tag == "getProperties":
                    # either no devicename, or an unknown device
                    # if it were a known devicename the previous block would have handled it.
                    # so send it on all connections
                    remcon.send(xmldata)
                else:
                    # Check if this remcon is snooping on this device/vector
                    if remcon.clientdata["snoopall"]:
                        remcon.send(xmldata)
                    elif devicename and (devicename in remcon.clientdata["snoopdevices"]):
                        remcon.send(xmldata)
                    elif devicename and propertyname and ((devicename, propertyname) in remcon.clientdata["snoopvectors"]):
                        remcon.send(xmldata)

            # transmit xmldata out to other drivers
            for driver in self.alldrivers:
                if driver is self.driver:
                    continue
                if xmldata.tag == "getProperties":
                    # either no devicename, or an unknown device
                    await driver.readerque.put(xmldata)
                else:
                    # Check if this driver is snooping on this device/vector
                    if driver.snoopall:
                        await driver.readerque.put(xmldata)
                    elif devicename and (devicename in driver.snoopdevices):
                        await driver.readerque.put(xmldata)
                    elif devicename and propertyname and ((devicename, propertyname) in driver.snoopvectors):
                        await driver.readerque.put(xmldata)


            # traffic from this driver writerque has been sent to other drivers/remotes if they want to snoop.
            # The traffic must also now be sent to the clients.
            # If no clients are connected, do not put this data into
            # the serverwriterque
            for clientconnection in self.connectionpool:
                if clientconnection.connected:
                    # at least one is connected, so this data is put into
                    # serverwriterque, and is then sent to each client by
                    # the _sendtoclient method.
                    await self.serverwriterque.put(xmldata)
                    break
            # task completed
            writerque.task_done()


class _ClientConnection:

    "Handles a client connection"

    def __init__(self, devices, exdrivers, remotes, serverreaderque):
        # self.txque will have data to be transmitted
        # inserted into it from the IPyServer._sendtoclient()
        # method
        self.txque = asyncio.Queue(6)

        # devices is a dictionary of device name to device
        self.devices = devices
        self.remotes = remotes
        self.exdrivers = exdrivers
        self.serverreaderque = serverreaderque
        # self.connected is True if this pool object is running a connection
        self.connected = False

        # timer used to force a data transmission after timeout seconds
        # this will cause an exception if the connection is broken and will shut down
        # the connection
        self.txtimer = TXTimer(timeout = 15)
        self.rxtimer = TXTimer(timeout = 15)

        self.rx = None
        self.tx = None

        self._stop = False       # Gets set to True to stop communications

    @property
    def stop(self):
        "returns self._stop"
        return self._stop

    def shutdown(self):
        "Sets self.stop to True and calls shutdown on tasks"
        self._stop = True
        self.connected = False
        if not self.rx is None:
            self.rx.shutdown()
        if not self.tx is None:
            self.tx.shutdown()

    async def _monitor_connection(self):
        """If connected, send def vectors every timeout seconds
           This ensures that if the connection has failed, due to the client disconnecting, the write
           to the port operation will cause a failure exception which will close the connection"""
        # this only operates if there are local ipydrivers connected
        if not self.devices:
            return
        while not self._stop:
            await asyncio.sleep(5)
            # this is tested every five seconds
            # If a remcon is connected, leave the send def vectors to the remcon
            # by increasing the timeout, so the remcon timer times out first
            if self.remotes:
                for remcon in self.remotes:
                    if remcon.connected:
                        self.txtimer.timeout = 25
                        self.rxtimer.timeout = 25
                        break
                else:
                    self.txtimer.timeout = 15
                    self.rxtimer.timeout = 15
            if self.connected and self.txtimer.elapsed() and self.rxtimer.elapsed():
                # no transmission in timeout seconds so send defVectors
                for device in self.devices.values():
                    if not device.enable:
                        continue
                    for vector in device.values():
                        if not vector.enable:
                            continue
                        xmldata =  vector._make_defVector()
                        if xmldata:
                            await self.txque.put(xmldata)


    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"
        self.connected = True
        sendchecker = SendChecker(self.devices, self.exdrivers, self.remotes)
        addr = writer.get_extra_info('peername')
        self.rx = Port_RX(sendchecker, reader, self.rxtimer)
        self.tx = Port_TX(sendchecker, writer, self.txtimer)
        logger.info(f"Connection received from {addr}")
        try:
            txtask = asyncio.create_task(self.tx.run_tx(self.txque))
            rxtask = asyncio.create_task(self.rx.run_rx(self.serverreaderque))
            montask = asyncio.create_task(self._monitor_connection())
            await asyncio.gather(txtask, rxtask, montask)
        except ConnectionError:
            self.connected = False
            txtask.cancel()
            rxtask.cancel()
            montask.cancel()
            cleanque(self.txque)
        logger.info(f"Connection from {addr} closed")





import collections, asyncio, sys, copy

from datetime import datetime, timezone

import xml.etree.ElementTree as ET

import logging
logger = logging.getLogger(__name__)

from .comms import STDINOUT, Portcomms
from . import events
from .propertyvectors import timestamp_string


class IPyDriver(collections.UserDict):

    """A subclass of this should be created with methods written
       to control your device.

       devices is a list of Device objects this driver handles.

       coros are optional positional arguments, being co-routines that
       you may have created to operate your instruments, the co-routines set
       here will all be awaited when the driver is run.

       You may optionally include named arguments of any hardware
       data that may be usefull to you, these will be available as
       attribute dictionary self.driverdata.

       This object is also a mapping, of devicename:deviceobject
       """

    @staticmethod
    def indi_number_to_float(value):
        """The INDI spec allows a number of different number formats, given any number string, this returns a float.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        try:
            if isinstance(value, float):
                return value
            if isinstance(value, int):
                return float(value)
            if not isinstance(value, str):
                raise TypeError
            # negative is True, if the value is negative
            value = value.strip()
            negative = value.startswith("-")
            if negative:
                value = value.lstrip("-")
            # Is the number provided in sexagesimal form?
            if value == "":
                parts = ["0", "0", "0"]
            elif " " in value:
                parts = value.split(" ")
            elif ":" in value:
                parts = value.split(":")
            elif ";" in value:
                parts = value.split(";")
            else:
                # not sexagesimal
                parts = [value, "0", "0"]
            if len(parts) > 3:
                raise TypeError
            # Any missing parts should have zero
            if len(parts) == 1:
                parts.append("0")
                parts.append("0")
            if len(parts) == 2:
                parts.append("0")
            assert len(parts) == 3
            # a part could be empty string, ie if 2:5: is given
            numbers = list(float(x) if x else 0.0 for x in parts)
            floatvalue = numbers[0] + (numbers[1]/60) + (numbers[2]/3600)
            if negative:
                floatvalue = -1 * floatvalue
        except Exception:
            raise TypeError("Error: Unable to parse number value")
        return floatvalue


    def __init__(self, devices, *coros, **driverdata):
        super().__init__()

        # devices is either a Device, or a list of Device objects
        if isinstance(devices, Device):
            self.devices = {devices.devicename:devices}
        else:
            # this is a dictionary of device name to device this driver owns
            self.devices = {d.devicename:d for d in devices}

        for device in self.devices.values():
            # set driver into devices
            device.driver = self

        self.data = self.devices

        # self.data is used by UserDict, it is an alias of self.devices
        # simply because 'devices' is more descriptive

        # _coros is a set of co_routines to be started
        self._coros = coros
        # self._tasks holds tasks to be generated by self.asyncrun()
        self._tasks = []

        # dictionary of optional data
        self.driverdata = driverdata

        # traffic is transmitted out on the writerque
        self.writerque = asyncio.Queue(4)
        # and read in from the readerque
        self.readerque = asyncio.Queue(4)
        # and snoop data is passed on to the snoopque
        self.snoopque = asyncio.Queue(4)
        # data for each device is passed to each device dataque

        # An object for communicating can be set, if not set, then
        # self.comms = STDINOUT() will be set in the asyncrun call
        self.comms = None

        # These set the remote traffic which this driver is snooping
        # initially the driver is not snooping anything, until it sends
        # a getProperties
        self.snoopall = False           # gets set to True if it is snooping everything
        self.snoopdevices = set()       # gets set to a set of device names
        self.snoopvectors = set()       # gets set to a set of (devicename,vectorname) tuples

        # If True, xmldata will be logged at DEBUG level
        self.debug_enable = True

        # shutdown routine sets this to True to stop coroutines
        self._stop = False

    def shutdown(self):
        "Shuts down the driver, sets the flag self._stop to True"
        self._stop = True
        if not self.comms is None:
            self.comms.shutdown()
        for device in self.devices.values():
            device.shutdown()

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop


    def listen(self, host="localhost", port=7624):
        """If called, sets up listening on the given host and port.
           Only one connection will accepted, further connection attempts
           while a client is already connected will be refused.
           This method also checks for enableBLOB instructions, and implements them.
           In general, using IPyServer is preferred."""
        if not self.comms is None:
             raise RuntimeError("A communications method has already been set, there can only be one")
        self.comms = Portcomms(self.devices, host, port)


    async def send(self, xmldata):
        "Transmits xmldata, this is an internal method, not normally called by a user."
        if self.comms.connected:
            await self.writerque.put(xmldata)
            if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
                if (xmldata.tag == "setBLOBVector") and len(xmldata):
                    data = copy.deepcopy(xmldata)
                    for element in data:
                        element.text = "NOT LOGGED"
                    binarydata = ET.tostring(data)
                    logger.debug(f"TX:: {binarydata.decode('utf-8')}")
                else:
                    binarydata = ET.tostring(xmldata)
                    logger.debug(f"TX:: {binarydata.decode('utf-8')}")


    def __setitem__(self, devicename):
        raise KeyError

    async def _read_readerque(self):
        client_tags = ("enableBLOB", "newSwitchVector", "newNumberVector", "newTextVector", "newBLOBVector")
        snoop_tags = ("message", 'delProperty', 'defSwitchVector', 'setSwitchVector', 'defLightVector',
                      'setLightVector', 'defTextVector', 'setTextVector', 'defNumberVector', 'setNumberVector',
                      'defBLOBVector', 'setBLOBVector')
        while not self._stop:
            await asyncio.sleep(0)
            # reads readerque, and sends xml data to the device via its dataque
            root = await self.readerque.get()
            # log the received data
            if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
                if ((root.tag == "setBLOBVector") or (root.tag == "newBLOBVector")) and len(root):
                    data = copy.deepcopy(root)
                    for element in data:
                        element.text = "NOT LOGGED"
                    binarydata = ET.tostring(data)
                    logger.debug(f"RX:: {binarydata.decode('utf-8')}")
                else:
                    binarydata = ET.tostring(root)
                    logger.debug(f"RX:: {binarydata.decode('utf-8')}")
            if root.tag == "getProperties":
                version = root.get("version")
                if version != "1.7":
                    self.readerque.task_done()
                    continue
                # getProperties received with correct version
                devicename = root.get("device")
                # devicename is None (for all devices), or a named device
                if devicename is None:
                    for d in self.devices.values():
                        if d.enable:
                            await d.dataque.put(root)
                elif devicename in self.devices:
                    if self.devices[devicename].enable:
                        await self.devices[devicename].dataque.put(root)
                # else device not recognised
            elif root.tag in client_tags:
                # xml received from client
                devicename = root.get("device")
                if devicename is None:
                    # device not given, ignore this
                    self.readerque.task_done()
                    continue
                elif devicename in self.devices:
                    if self.devices[devicename].enable:
                        await self.devices[devicename].dataque.put(root)
                # else device not recognised
            elif root.tag in snoop_tags:
                # xml received from other devices
                await self.snoopque.put(root)
            self.readerque.task_done()


    async def _snoophandler(self):
        """Creates events using data from self.snoopque"""
        while not self._stop:
            # get block of data from the self.snoopque
            await asyncio.sleep(0)
            root = await self.snoopque.get()
            devicename = root.get("device")
            if devicename is not None:
                # if a device name is given, check
                # it is not in this drivers devices
                if devicename in self.devices:
                    logger.error("Cannot snoop on a device already controlled by this driver")
                    self.snoopque.task_done()
                    continue
            try:
                if root.tag == "message":
                    # create event
                    event = events.message(root)
                    await self.snoopevent(event)
                elif root.tag == "delProperty":
                    # create event
                    event = events.delProperty(root)
                    await self.snoopevent(event)
                elif root.tag == "defSwitchVector":
                    # create event
                    event = events.defSwitchVector(root)
                    await self.snoopevent(event)
                elif root.tag == "setSwitchVector":
                    # create event
                    event = events.setSwitchVector(root)
                    await self.snoopevent(event)
                elif root.tag == "defLightVector":
                    # create event
                    event = events.defLightVector(root)
                    await self.snoopevent(event)
                elif root.tag == "setLightVector":
                    # create event
                    event = events.setLightVector(root)
                    await self.snoopevent(event)
                elif root.tag == "defTextVector":
                    # create event
                    event = events.defTextVector(root)
                    await self.snoopevent(event)
                elif root.tag == "setTextVector":
                    # create event
                    event = events.setTextVector(root)
                    await self.snoopevent(event)
                elif root.tag == "defNumberVector":
                    # create event
                    event = events.defNumberVector(root)
                    await self.snoopevent(event)
                elif root.tag == "setNumberVector":
                    # create event
                    event = events.setNumberVector(root)
                    await self.snoopevent(event)
                elif root.tag == "defBLOBVector":
                    # create event
                    event = events.defBLOBVector(root)
                    await self.snoopevent(event)
                elif root.tag == "setBLOBVector":
                    # create event
                    event = events.setBLOBVector(root)
                    await self.snoopevent(event)
            except events.EventException as ex:
                # if an EventException is raised, it is because received data is malformed
                # so log it
                logger.exception("An exception occurred creating a snoop event")
            self.snoopque.task_done()

    async def send_message(self, message="", timestamp=None):
        "Send system wide message - without device name"
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_message must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('message')
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.send(xmldata)

    async def send_getProperties(self, devicename=None, vectorname=None):
        """Sends a getProperties request - which is used to snoop data from other devices
           on the network, if devicename given, it must not be a device of this driver as
           the point of this is to snoop on remote devices."""
        xmldata = ET.Element('getProperties')
        xmldata.set("version", "1.7")
        if devicename is None:
            await self.send(xmldata)
            self.snoopall = True
            return
        if devicename in self.devices:
            logger.error("Cannot snoop on a device already controlled by this driver")
            return
        xmldata.set("device", devicename)
        if vectorname is None:
            await self.send(xmldata)
            self.snoopdevices.add(devicename)
            return
        xmldata.set("name", vectorname)
        await self.send(xmldata)
        # adds tuple (devicename,vectorname) to self.snoopvectors
        self.snoopvectors.add((devicename,vectorname))


    async def hardware(self):
        """Override this to operate device hardware, and transmit updates

        For example: call your own code to operate hardware
        then update the appropriate vectors, and send updated
        values to the client using
        await vector.send_setVector()"""
        pass


    async def rxevent(self, event):
        """Override this. On receiving data, this is called, and should
           handle any necessary actions.
           event is an object describing the event, with attributes
           devicename, vectorname, vector, root
           where vector is the properties vector the event refers to, and
           root is an xml.etree.ElementTree object of the received xml"""
        match event:
            case events.getProperties():
                await event.vector.send_defVector()


    async def snoopevent(self, event):
        """Override this if this driver is snooping on other devices.
           On receiving snoop data, this is called, and should handle
           any necessary actions.
           event is an object with attributes according to the data received."""
        pass

    async def asyncrun(self):
        """Gathers tasks to be run simultaneously"""

        logger.info(f"Driver {self.__class__.__name__} started")

        # set an object for communicating, as default this is stdin and stdout
        if self.comms is None:
            self.comms = STDINOUT()

        # get all tasks into a list

        self._tasks.append( self.comms(self.readerque, self.writerque) )    # run communications
        self._tasks.append( self.hardware() )                               # task to operate device hardware, and transmit updates
        self._tasks.append( self._read_readerque() )                        # task to handle received xml data
        self._tasks.append( self._snoophandler() )                          # task to handle incoming snoop data


        for device in self.devices.values():
            self._tasks.append(device._handler())                           # each device handles its incoming data
            for pv in device.propertyvectors.values():
                self._tasks.append(pv._handler())                           # each property handles its incoming data
                # also give the propertyvector a reference to this driver
                # so it can call eventaction and have access to writerque
                pv.driver = self

        await asyncio.gather( *self._coros, *self._tasks )


class Device(collections.UserDict):

    """An instance of this should be created for each device controlled by this driver.
       The properties argument is a list of vectors controlling this device.
       devicedata will be an attribute dictionary of any hardware data that may be usefull.
    """

    def __init__(self, devicename, properties, **devicedata):
        super().__init__()

        # This device name
        self.devicename = devicename

        # if self.enable is False, this device ignores incoming traffic
        # and (apart from delProperty) does not transmit anything
        # from his device
        self.enable = True

        # the driver places data in this que to send data to this device
        self.dataque = asyncio.Queue(4)

        # Every property of this device has a dataque, which is set into this dictionary
        self.propertyquedict = {p.name:p.dataque for p in properties}

        # dictionary of optional data
        self.devicedata = devicedata

        # this will be set when the driver asyncrun is run
        self.driver = None

        # this is a dictionary of property name to propertyvector this device owns
        self.propertyvectors = {}
        for p in properties:
            p.devicename = self.devicename
            self.propertyvectors[p.name] = p

        self.data = self.propertyvectors
        # self.data is used by UserDict, it is an alias of self.propertyvectors
        # simply because 'propertyvectors' is more descriptive

       # shutdown routine sets this to True to stop coroutines
        self._stop = False

    def shutdown(self):
        """Shuts down the device, sets the flag self._stop to True
           and shuts down property vector handlers"""
        self._stop = True
        for pv in self.propertyvectors.values():
            pv.shutdown()

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    async def send_device_message(self, message="", timestamp=None):
        """Send a message associated with this device, which the client could display.
           The timestamp should be either None or a datetime.datetime object. If the
           timestamp is None a UTC value will be inserted."""
        if not self.enable:
            # messages only sent if self.enable is True
            return
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_device_message must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('message')
        xmldata.set("device", self.devicename)
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.driver.send(xmldata)

    async def send_delProperty(self, message="", timestamp=None):
        """Sending delProperty with this device method, (as opposed to the vector send_delProperty method)
           informs the client this device is not available, it also sets a device.enable attribute to
           False, which stops any data being transmitted between the client and this device.
           Setting device.enable to True re-enables communications.
           The message argument is any appropriate string which the client could display to the user.
           The timestamp should be either None or a datetime.datetime object. If the timestamp is None
           a UTC value will be inserted."""
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_delProperty must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('delProperty')
        xmldata.set("device", self.devicename)
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.driver.send(xmldata)
        self.enable = False


    async def devhardware(self, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'hardware' method if it is used."""
        pass


    async def devrxevent(self, event, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'rxevent' method if it is used."""
        pass


    async def devsnoopevent(self, event, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'snoopevent' method if it is used."""
        pass


    def __setitem__(self, vectorname):
        raise KeyError

    async def _handler(self):
        """Handles data read from dataque"""
        while not self._stop:
            await asyncio.sleep(0)
            # get block of data from the self.dataque
            root = await self.dataque.get()

            if not self.enable:
                self.dataque.task_done()
                continue
            if root.tag == "getProperties":
                name = root.get("name")
                # name is None (for all properties), or a named property
                if name is None:
                    for pvector in self.propertyvectors.values():
                        if pvector.enable:
                            await pvector.dataque.put(root)
                elif name in self.propertyvectors:
                    if self.propertyvectors[name].enable:
                        await self.propertyvectors[name].dataque.put(root)
                else:
                    # property name not recognised
                    self.dataque.task_done()
                    continue
            elif root.tag == "enableBLOB":
                name = root.get("name")
                # name is None (for all properties), or a named property
                if name is None:
                    for pvector in self.propertyvectors.values():
                        if pvector.enable:
                            await pvector.dataque.put(root)
                elif name in self.propertyvectors:
                    if self.propertyvectors[name].enable:
                        await self.propertyvectors[name].dataque.put(root)
                else:
                    # property name not recognised
                    self.dataque.task_done()
                    continue
            else:
                # root.tag will be one of
                # newSwitchVector, newNumberVector, newTextVector, newBLOBVector
                name = root.get("name")
                if name is None:
                    # name not given, ignore this
                    self.dataque.task_done()
                    continue
                elif name in self.propertyvectors:
                    pvector = self.propertyvectors[name]
                    if pvector.perm != "ro" and pvector.enable:
                        # all ok, add to the vector dataque
                        await pvector.dataque.put(root)
            # task completed
            self.dataque.task_done()
